import argparse
import inspect
import json
import pathlib
import sys

import miko

NO_ACTION = """\
usage: miko [-h] [--version] {info,clean} ...
miko: error: the following arguments are required: action"""


class MikoSignature:
    class Parameter:
        def __init__(self, annotation: str) -> None:
            annotation = str(annotation).strip()
            if "=" in annotation:
                annotation, _, self.default = annotation.partition("=")
                self.default = self.default.strip()
            else:
                self.default = inspect._empty

            if ":" in annotation:
                annotation, _, self.annotation = annotation.partition(":")
                self.annotation = self.annotation.strip()
            else:
                self.annotation = inspect._empty

            self.name = annotation.strip().replace(" ", "")

        def as_dict(self, camelCase: bool = False):
            return {
                "name": self.name,
                "annotation": self.annotation if self.annotation is not inspect._empty else None,
                "default": self.default if self.default is not inspect._empty else None
            }

    def __init__(self, definition: str, line: int) -> None:
        self.line = int(line)

        definition = str(definition).strip()
        _, _, definition = definition.partition("(")
        annotations, _, definition = definition.partition(")")

        _, _, definition = definition.rpartition("->")
        definition = definition.strip(":")
        self.return_annotation = str(definition).strip()
        self.return_annotation = self.return_annotation if self.return_annotation else inspect._empty

        annotations_results = [""]

        OPENED_BRACKET = 0

        for c in annotations:
            if c == "," and OPENED_BRACKET <= 0:
                annotations_results.append("")
                continue
            elif c == "[":
                OPENED_BRACKET += 1
            elif c == "]":
                OPENED_BRACKET -= 1

            annotations_results[-1] += c

        # print("annotations", annotations)
        # print("annotations_results", annotations_results)

        results = [self.Parameter(v) for v in annotations_results if v]  # problem if using type annotation with commas like Union[hello, world]
        self.parameters = {v.name: v for v in results}

    def as_dict(self, camelCase: bool = False):
        return {
            "line": self.line,
            "parameters": {v: k.as_dict(camelCase) for v, k in self.parameters.items()},
            "returnAnnotation" if camelCase else "return_annotation": self.return_annotation if self.return_annotation is not inspect._empty else None
        }


class FileReadingElement:
    def __init__(self, start_line: int = None, end_line: int = None, indent: int = 0, docstring: str = "", signature: MikoSignature = None, had_docstring: bool = False, quotation: str = '"""', noself: bool = False) -> None:
        self.start_line = start_line
        self.end_line = end_line
        self.indent = indent
        self.docstring = docstring
        self.signature = signature
        self.had_docstring = had_docstring
        self.quotation = str(quotation)
        self.noself = noself

    @property
    def docs(self) -> miko.Docs:
        return miko.Docs(self.docstring, self.signature, noself=self.noself)

    def as_dict(self, camelCase: bool = False):
        return {
            "startLine" if camelCase else "start_line": self.start_line,
            "endLine" if camelCase else "end_line": self.end_line,
            "indent": self.indent,
            "docs": self.docs.as_dict(camelCase),
            "signature": self.signature.as_dict(camelCase) if self.signature else None,
            "hadDocstring" if camelCase else "had_docstring": self.had_docstring,
            "quotation": self.quotation
        }


def read_file(text: str, noself: bool = False):
    results = []

    IN_DOCSTRING = False
    LAST_ELEMENT = FileReadingElement(noself=noself)
    for index, line in enumerate(text.splitlines(), start=1):
        current_indent = len(line) - len(line.lstrip())
        docstring_line = line.removeprefix(" " * LAST_ELEMENT.indent)
        line = line.strip()

        if LAST_ELEMENT.signature is not None and index - LAST_ELEMENT.signature.line > 1 and not IN_DOCSTRING:
            # print("LINE:", line, "HAS_SIGNATURE_AND_PASSED", index - LAST_ELEMENT.signature.line)
            results.append(LAST_ELEMENT)
            LAST_ELEMENT = FileReadingElement(noself=noself)

        if line.startswith("def ") and not IN_DOCSTRING:
            # print("LINE:", line, "STARTSWITH_DEF")
            LAST_ELEMENT.signature = MikoSignature(line, index)
            LAST_ELEMENT.start_line = index
            LAST_ELEMENT.indent = current_indent + 4
        elif (line.startswith('"""') or line.startswith("'''")):
            if line.startswith('"""'):
                if '"""' in line.removeprefix('"""'):  # if inline
                    continue
            else:
                if "'''" in line.removeprefix("'''"):  # if inline
                    continue
            if IN_DOCSTRING and line.startswith(LAST_ELEMENT.quotation):
                # print("LINE:", line, "IN_DOCSTRING_STARTSWITH_QUOTATION:", LAST_ELEMENT.quotation)
                LAST_ELEMENT.end_line = index
                results.append(LAST_ELEMENT)
                LAST_ELEMENT = FileReadingElement(noself=noself)
                IN_DOCSTRING = False
            elif not IN_DOCSTRING:
                # print("LINE:", line, "NOT_IN_DOCSTRING")
                if not LAST_ELEMENT.signature:
                    LAST_ELEMENT.start_line = index
                LAST_ELEMENT.indent = current_indent
                LAST_ELEMENT.had_docstring = True
                LAST_ELEMENT.quotation = '"""' if line.startswith('"""') else "'''"
                IN_DOCSTRING = True
                LAST_ELEMENT.docstring += line.removeprefix(LAST_ELEMENT.quotation) + "\n"
        elif IN_DOCSTRING:
            # print("LINE:", line, "IN_DOCSTRING")
            LAST_ELEMENT.docstring += docstring_line + "\n"

    if not LAST_ELEMENT.had_docstring and LAST_ELEMENT.signature is not None:
        # print("LINE:", "HAD_DOCSTRING_AND_SIGNATURE")
        results.append(LAST_ELEMENT)

    return results


def main():
    parser = argparse.ArgumentParser("miko", description="See how to use a Python object at a glance!")
    parser.add_argument('--version', '-v', action='version', version=miko.__version__)

    subparser = parser.add_subparsers(help='Actions', dest="action")

    parser_info = subparser.add_parser('info', help='Retrieve info on the given docstring')
    parser_info.add_argument('--text', '-t', action='store', type=str, required=False, help='The docstring to get the information from')
    parser_info.add_argument("--file", "-f", action='store', type=str, required=False, help='The file to get the docstrings from.')
    parser_info.add_argument("--indent", "-i", action='store', type=int, required=False, default=4, help='The indentation for the JSON result.')
    parser_info.add_argument("--noself", action='store_true', required=False, default=False,
                             help='Ignoring the "self" parameter from signatures. (useful for class methods)')

    parser_clean = subparser.add_parser("clean", help="Clean the docstring")
    parser_clean.add_argument('--text', '-t', action='store', type=str, required=False, help='The docstring to clean')
    parser_clean.add_argument("--file", "-f", action='store', type=str, required=False, help='The file to get the docstrings from.')
    parser_clean.add_argument("--output", "-o", action='store', type=str, required=False,
                              default=None, help='The file to output the cleaned result to.')

    parser_clean.add_argument("--indent", "-i", action='store', type=int, required=False, default=4, help='The indentation to clean the docs.')
    parser_clean.add_argument("--noself", action='store_true', required=False, default=False,
                              help='Ignoring the "self" parameter from signatures. (useful for class methods)')

    args = parser.parse_args()

    if not args.action:
        print(NO_ACTION)
        sys.exit(1)

    if args.text:
        if args.action == "clean":
            result = miko.Docs(args.text, noself=args.noself).dumps(indent=args.indent)
            if parser_clean.output:
                with open(parser_clean.output, "w") as f:
                    f.write(result)
                return
            return print(result)
        return print(json.dumps(miko.Docs(args.text, noself=args.noself).as_dict(True), indent=args.indent, ensure_ascii=False))

    file = pathlib.Path(args.file)
    if not file.is_file():
        raise ValueError("The given file '{path}' does not seem to be a file".format(path=file))

    with open(file) as f:
        text = f.read()
    # text = file.read_text()

    results = read_file(text, noself=args.noself)
    if args.action == "clean":
        returning = []
        INDEXES = {}
        for element in results:
            # element: FileReadingElement
            # if element.signature:
            INDEXES[element.start_line] = element
            if element.end_line:
                for l in range(element.start_line + 1, element.end_line + 1):
                    INDEXES[l] = False
        # {
        #     "line<int>": FileReadingElement,
        #     "line+1<int>": False,
        #     "line+2<int>": False,

        #     "line<int>": FileReadingElement,
        #     "line+1<int>": False,
        # }
        for index, line in enumerate(text.splitlines(), start=1):
            current = INDEXES.get(index, None)

            if current is False:
                # print("🈲 LINE", str(index).zfill(3), ":", '\033[91m', line, '\033[0m')
                continue
            # print("✅ LINE", str(index).zfill(3), ":", '\033[92m', line, '\033[0m')

            if current is None or current.signature:
                returning.append(line)

            if current is None:
                continue

            # current: FileReadingElement

            returning.append("{i}{q}".format(i=" " * current.indent, q=current.quotation))
            for l in current.docs.dumps(indent=args.indent).splitlines():
                returning.append("{i}{l}".format(i=" " * current.indent, l=l))
            returning.append("{i}{q}".format(i=" " * current.indent, q=current.quotation))
        output = "\n".join(returning)
        if args.output:
            with open(args.output, "w") as f:
                f.write(output)
            return
        return print(output)
    print(json.dumps([r.as_dict(True) for r in results], indent=args.indent, ensure_ascii=False))


if __name__ == "__main__":
    main()



"""
Defines the base parser
"""
import inspect


class Parser:
    """
    The base class for the parsers
    """
    __map_attribute__ = "parser"

    def __init__(self, body: str = "", signature: inspect.Signature = None, **kwargs) -> None:
        """
        Parameters
        ----------
        body: str, default = ""
        signature: inspect.Signature, default = None
        **kwargs

        Returns
        -------
        None
        """
        self.original = str(body)
        self.signature = signature

    def as_dict(self, camelCase: bool = False):
        """
        Parameters
        ----------
        camelCase: bool, default = False
        """
        return {
            "original": self.original
        }



import inspect
import typing

from miko.parser.parser import Parser


class ListElement:
    def __init__(self, name: str, options: list = None, content: list = None, signature: inspect.Signature = None) -> None:
        self.name = name.__name__ if isinstance(name, type) else str(name)
        self.options = [v for v in options if v] if options else []
        self.content = [c for c in content if c] if content else []
        self.signature = signature

    def __repr__(self) -> str:
        return "<ListElement options={options}, content={content} lines>".format(options=", ".join(self.options) if len(self.options) > 1 else "None", content=len(self.content))

    def as_dict(self, camelCase: bool = False):
        return {
            "name": self.name,
            "options": self.options,
            "content": self.content
        }


class List(Parser):
    __element_type__ = ListElement
    __map_attribute__ = "list"

    def __init__(self, body: str = "", signature: inspect.Signature = None, **kwargs) -> None:
        super().__init__(body, signature)
        self.elements = {}
        self.extend(self.original, add_to_original=False)

    def __iter__(self) -> typing.Iterator:
        """
        Returns the iterator.
        """
        return self.elements.values().__iter__()

    def __len__(self):
        return len(self.elements)

    def as_dict(self, camelCase: bool = False):
        results = super().as_dict(camelCase)
        results.update({
            "elements": {str(k) if camelCase else k: v.as_dict(camelCase) for k, v in self.elements.items()}
        })
        return results

    def extend(self, content: str, add_to_original: bool = True):
        current = None
        for line in content.splitlines():
            if line.startswith(" "):
                if current is None:
                    continue
                line = line.strip()
                self.elements[current].content.append(line)
                continue
            name, _, options = line.partition(":")
            current, options = [v.strip() for v in (name, options)]
            if current not in self.elements:
                self.elements[current] = self.__element_type__(name=current, options=[opt.strip()
                                                               for opt in options.split(",")], signature=self.signature)
            else:
                self.elements[current].options.extend([opt.strip() for opt in options.split(",")])

        if add_to_original:
            self.original += "\n" + content

    def __repr__(self) -> str:
        return "<{name} elements={elements}>".format(name=self.__class__.__name__, elements=list(self.elements.keys()))

    def __getattr__(self, key: str):
        return self.elements.__getitem__(key)


class Parameter(ListElement):
    @property
    def deprecated(self) -> bool:
        return "deprecated" in [str(v).lower() for v in self.options]

    @property
    def optional(self) -> bool:
        options = [str(v).lower() for v in self.options]
        return "optional" in options or any(v.startswith("default") for v in options)

    @property
    def default(self) -> typing.Optional[str]:
        for opt in self.options:
            opt = str(opt)
            if opt.lower().startswith("default"):
                _, _, content = opt.partition("=")
                element = content.strip()
                # i could parse some types here
                return element

        try:
            result = self.signature.parameters[self.name].default
            if not isinstance(result, inspect._empty) and not result is inspect._empty:
                return result.__name__ if isinstance(result, type) else str(result)
        except Exception:
            return None

    @property
    def types(self) -> set:
        results = set()
        for opt in self.options:
            option = str(opt).lower()
            if option.startswith("default") or option in {"optional", "required", "deprecated"}:
                continue
            results.update([v.strip() for v in str(opt).split("|") if str(v).strip().replace(" ", "") != "<class'inspect._empty'>"])

        try:
            param = self.signature.parameters[self.name]
            annotation = param.annotation
            if annotation is not inspect._empty and str(annotation).strip().replace(" ", "") != "<class 'inspect._empty'>":
                if hasattr(annotation, "__origin__") and annotation.__origin__ is typing.Union:
                    for a in annotation.__args__:
                        if str(a).lower().strip('"').strip("'").strip() not in {str(r).lower() for r in results}:
                            results.add(a)
                elif annotation:
                    if str(annotation).lower().strip('"').strip("'").strip() not in {str(r).lower() for r in results}:
                        results.add(annotation)
        except Exception:
            pass

        return results

    def __repr__(self) -> str:
        return "<Parameter types={types} optional={optional} default={default}>".format(types=self.types, optional=self.optional, default=self.default)

    def as_dict(self, camelCase: bool = False):
        results = super().as_dict(camelCase)
        results.update({
            "types": [str(t) for t in self.types] if camelCase else self.types,
            "optional": self.optional,
            "default": self.default,
            "deprecated": self.deprecated
        })
        return results


class Parameters(List):
    __element_type__ = Parameter
    __map_attribute__ = "parameters"

    def __init__(self, body: str = "", signature: inspect.Signature = None, noself: bool = False, **kwargs) -> None:
        super().__init__(body, signature)
        if signature:
            for parameter in signature.parameters:
                parameter = str(parameter)
                if parameter == "self" and noself:
                    continue
                if parameter not in self.elements:
                    self.elements[parameter] = self.__element_type__(name=parameter, signature=signature)


class Returns(List):
    __map_attribute__ = "returns"

    def __init__(self, body: str = "", signature: inspect.Signature = None, **kwargs) -> None:
        super().__init__(body, signature)
        if signature:
            annotation = signature.return_annotation
            if hasattr(annotation, "__origin__") and annotation.__origin__ is typing.Union:
                annotations = [str(v) for v in annotation.__args__]
            else:
                annotations = [annotation]
            for annotation in annotations:
                if annotation not in self.elements and not isinstance(annotation, inspect._empty) and not annotation is inspect._empty:
                    self.elements[annotation] = self.__element_type__(name=annotation)


class Raises(List):
    __map_attribute__ = "raises"


class Changelog(List):
    __map_attribute__ = "changelog"


class Copyright(List):
    __map_attribute__ = "copyright"



from os import path

from setuptools import setup

with open(path.join(path.abspath(path.dirname(__file__)), "README.md"), encoding="utf-8") as f:
    readme_description = f.read()


def read_requirements(filename):
    with open(filename, "r", encoding="utf-8") as fp:
        return fp.read().strip().splitlines()


setup(
    name="miko",
    packages=["miko"],
    version="1.0",
    license="MIT License",
    description="See how to use a Python object at a glance!",
    author="Anime no Sekai",
    author_email="niichannomail@gmail.com",
    url="https://github.com/Animenosekai/miko",
    download_url="https://github.com/Animenosekai/miko/archive/v1.0.tar.gz",
    keywords=[
        "python",
        "documentation",
        "miko",
        "docs"
    ],
    install_requires=read_requirements("requirements.txt"),
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    long_description=readme_description,
    long_description_content_type="text/markdown",
    include_package_data=True,
    python_requires=">=3.4, <4",
    entry_points={"console_scripts": ["miko = miko.__main__:main"]},
    package_data={
        "miko": ["LICENSE"],
    },
)



"""
Miko
A new Python documentation style

Author
------
Animenosekai
    Original author
"""

from miko.miko import Function, Docs


__version_tuple__ = (1, 0)


def __version_string__():
    if isinstance(__version_tuple__[-1], str):
        return '.'.join(map(str, __version_tuple__[:-1])) + __version_tuple__[-1]
    return '.'.join(str(i) for i in __version_tuple__)


__author__ = 'Anime no Sekai'
__copyright__ = 'Copyright 2022, miko'
__credits__ = ['animenosekai']
__license__ = 'MIT License'
__version__ = 'miko v{}'.format(__version_string__())
__maintainer__ = 'Anime no Sekai'
__email__ = 'niichannomail@gmail.com'
__status__ = 'Stable'



"""
miko.py

Contains the main code for the Miko documentation style
"""

import inspect
import sys
import types
import typing

import miko.parser.example as example_parsers
import miko.parser.list as list_parsers


class Function:
    def __init__(self, func: typing.Callable) -> None:
        self.original = func
        self.name = self.original.__name__
        self.signature = inspect.signature(func)
        self.docs = Docs(
            docs=self.original.__doc__ if self.original.__doc__ is not None else "No description",
            signature=self.signature
        )
        self.code = Function.get_code(self.original)
        self.source_filename = self.code.co_filename
        self.source_linenumber = self.code.co_firstlineno
        self.source_name = self.code.co_name
        self.local_variables = self.code.co_names
        self.parameters = self.signature.parameters
        self.return_annotation = self.signature.return_annotation

    @staticmethod
    def get_code(obj: typing.Any) -> types.CodeType:
        """
        Returns the __code__ object of a given callable object.

        Parameters
        ----------
        obj: Callable
            The object to get the __code__ from

        Returns
        -------
        CodeType
            The __code__ object
        """
        if inspect.ismethod(obj):
            obj = obj.__func__
        if inspect.isfunction(obj):
            obj = obj.__code__
        if inspect.istraceback(obj):
            obj = obj.tb_frame
        if inspect.isframe(obj):
            obj = obj.f_code
        if not inspect.iscode(obj):
            raise TypeError('''method, function, traceback, frame,\
                or code object was expected, got {}'''.format(type(obj).__name__))
        return obj

    def __repr__(self) -> str:
        return "<Function '{name}'>".format(name=self.name)


SECTIONS_MAP = {
    "PARAMETERS": list_parsers.Parameters,
    "PARAMETER": list_parsers.Parameters,
    "PARAMS": list_parsers.Parameters,
    "PARAM": list_parsers.Parameters,
    "ARGUMENTS": list_parsers.Parameters,
    "ARGUMENT": list_parsers.Parameters,
    "ARGS": list_parsers.Parameters,
    "ARG": list_parsers.Parameters,

    "RETURNS": list_parsers.Returns,
    "RETURN": list_parsers.Returns,
    "RETURNING": list_parsers.Returns,

    "RAISES": list_parsers.Raises,
    "RAISE": list_parsers.Raises,
    "RAISING": list_parsers.Raises,
    "EXCEPTIONS": list_parsers.Raises,
    "EXCEPTION": list_parsers.Raises,
    "ERRORS": list_parsers.Raises,
    "ERROR": list_parsers.Raises,

    "CHANGELOG": list_parsers.Changelog,
    "CHANGES": list_parsers.Changelog,

    "COPYRIGHTS": list_parsers.Copyright,
    "COPYRIGHT": list_parsers.Copyright,
    "AUTHORS": list_parsers.Copyright,
    "AUTHOR": list_parsers.Copyright,

    "EXAMPLES": example_parsers.Example,
    "EXAMPLE": example_parsers.Example
}


class Docs:
    parameters = list_parsers.Parameters()
    returns = list_parsers.Returns()
    raises = list_parsers.Raises()
    changelog = list_parsers.Changelog()
    copyright = list_parsers.Copyright()
    example = example_parsers.Example()
    if sys.version_info >= (3, 6):
        warnings: typing.List[str]  # novermin
        notes: typing.List[str]  # novermin
    else:
        warnings = []
        notes = []

    __elements_mapping__ = {v.__map_attribute__ for v in (list_parsers.Parameters, list_parsers.Returns,
                                                          list_parsers.Raises, list_parsers.Changelog,
                                                          list_parsers.Copyright, example_parsers.Example)}

    def __init__(self, docs: str, signature: inspect.Signature = None, noself: bool = False) -> None:
        if docs is None:
            docs = ""
        self.original = inspect.cleandoc(str(docs))
        self.elements = {}

        self.warnings = []
        self.notes = []

        missing = {v.__map_attribute__: v for v in (list_parsers.Parameters, list_parsers.Returns,
                                                    list_parsers.Raises, list_parsers.Changelog,
                                                    list_parsers.Copyright, example_parsers.Example)}

        self.description = []

        for section in self.original.split("\n\n"):
            name, _, body = section.partition("\n---")
            if not body:
                self.description.append(name)  # name would be the content here
                continue
            parse = SECTIONS_MAP.get(name.replace(" ", "").upper().strip(), None)
            if parse is None:
                self.description.append("\n---".join((name, body)))
                continue
            content = body.lstrip("-").lstrip("\n")
            try:
                self.elements[parse.__map_attribute__].extend(content)
            except Exception:
                self.elements[parse.__map_attribute__] = parse(content, signature=signature, noself=noself)
                missing.pop(parse.__map_attribute__, None)
        self.original_sections = list(self.elements.keys())

        if signature:
            if list_parsers.Parameters.__map_attribute__ not in self.elements:
                self.elements[list_parsers.Parameters.__map_attribute__] = list_parsers.Parameters(signature=signature, noself=noself)
                missing.pop(list_parsers.Parameters.__map_attribute__, None)

            if list_parsers.Returns.__map_attribute__ not in self.elements:
                self.elements[list_parsers.Returns.__map_attribute__] = list_parsers.Returns(signature=signature)
                missing.pop(list_parsers.Returns.__map_attribute__, None)

        for attr, parse in missing.items():
            self.elements[attr] = parse()

        # HANDLING TAGS
        for index, line in enumerate(self.description):
            start, _, content = str(line).partition(":")
            start = start.replace(" ", "").upper().strip()
            content = content.strip()
            if start in {"WARNING", "WARNINGS"}:
                self.warnings.append(content)
                self.description[index] = "Warning: {content}".format(content=content)
            elif start in {"NOTE", "NOTES", "SEEALSO", "INFORMATION"}:
                self.notes.append(content)
                self.description[index] = "Note: {content}".format(content=content)

        if len(self.description) > 0:
            checking = str(self.description[0]).replace(" ", "").upper()
            self.deprecated = checking.startswith("!DEPRECATED!") or checking.startswith(
                "!DEPRECATION!") or checking.startswith("!DEPRECATE!") or checking.startswith("!DEPRECATIONNOTICE!")
            if self.deprecated:
                index = self.description[0].find("!")
                second_index = index + self.description[0][index + 1:].find("!") + 2
                self.description[0] = "! DEPRECATED !" + self.description[0][second_index:]

        self.description = "\n\n".join(self.description)

    def __repr__(self) -> str:
        return "<Docs sections={sections}>".format(sections=self.original_sections)

    def __getattr__(self, key: str):
        if key in self.__elements_mapping__:
            return self.elements.__getitem__(key)
        raise AttributeError

    def dumps(self, indent=4):
        result = ""
        if self.description.replace(" ", ""):
            result += self.description
            result += "\n\n"
        sections = []
        for section in self.elements.values():
            if not isinstance(section, example_parsers.Example) and len(section) <= 0:
                continue

            current_section = ""

            if isinstance(section, example_parsers.Example):
                if not section.original:
                    continue
                current_section += section.__class__.__name__ + "\n"
                current_section += "-" * len(section.__class__.__name__) + "\n"

                current_section += section.original + "\n"
            else:
                current_section += section.__class__.__name__ + "\n"
                current_section += "-" * len(section.__class__.__name__) + "\n"

                for element in section:
                    # element = self.parameters.__element_type__()
                    current_section += element.name

                    if isinstance(element, list_parsers.Parameter):
                        options = []
                        if len(element.types) > 0:
                            options.append(" | ".join([v.__name__ if isinstance(v, type) else str(v) for v in element.types]))
                        if element.default:
                            options.append("default = {default}".format(default=element.default))
                        elif element.optional:
                            options.append("optional")
                        if element.deprecated:
                            options.append("deprecated")

                        if len(options) > 0:
                            current_section += ": " + ", ".join(options)

                    current_section += "\n"

                    for sentence in element.content:
                        current_section += " " * indent + sentence + "\n"

            sections.append(current_section)

        result += "\n".join(sections)

        if result.endswith("\n\n"):
            result = result.removesuffix("\n")

        return result

    def as_dict(self, camelCase: bool = False):
        results = {
            "description": self.description,
            "notes": self.notes,
            "warnings": self.warnings,
            "deprecated": self.deprecated
        }
        results.update({k: v.as_dict(camelCase) for k, v in self.elements.items()})
        return results
